﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Diagnostics;
using System.Drawing;
using System.Text;
using System.Threading;
using System.Threading.Tasks;
using System.Windows.Forms;

namespace AsyncCalcPiWithBackgroundWorkerSample {

  partial class AsyncCalcPiForm : Form {

        CancellationTokenSource tokenSource;
        CancellationToken token;

        public AsyncCalcPiForm() {
            InitializeComponent();
            tokenSource = new CancellationTokenSource();
            token = tokenSource.Token;
            SetUIReady();
        }

        void ShowProgressState(CalcPiUserState state)
        {
            try
            {
                // Display progress in UI
                this.resultsTextBox.Text = state.Pi;
                this.calcToolStripProgressBar.Maximum = state.TotalDigits;
                this.calcToolStripProgressBar.Value = state.DigitsSoFar;
            }
            catch (Exception e)
            {
                Console.WriteLine("{0} in ShowProgressState", e.Message);
                throw e;
            }
        }

        class CalcPiUserState {
      public readonly string Pi;
      public readonly int TotalDigits;
      public readonly int DigitsSoFar;

      public CalcPiUserState(string pi, int totalDigits, int digitsSoFar) {
            this.Pi = pi;
            this.TotalDigits = totalDigits;
            this.DigitsSoFar = digitsSoFar;
          }
        }


        TimeSpan TimePiThread(
            int digits, IProgress<CalcPiUserState> progress, CancellationToken token)
        {
            // Track start time
            DateTime start = DateTime.Now;

            CalcPiWithProgress(digits, progress, token);

            // Return elapsed time
            DateTime end = DateTime.Now;
            TimeSpan elapsed = end - start;
            return elapsed;
        }

        void ProgressTestCancel(
            IProgress<CalcPiUserState> progress, CancellationToken token, CalcPiUserState state)
        {
            // Report progress
            if (progress != null)
                progress.Report(state);
            token.ThrowIfCancellationRequested();
        }

        void CalcPiWithProgress(int digits, IProgress<CalcPiUserState> progress, CancellationToken token)
        {
            StringBuilder pi = new StringBuilder("3", digits + 2);

            // Report initial progress
            ProgressTestCancel(progress, token, new CalcPiUserState(pi.ToString(), digits, 0));

            if (digits > 0)
            {
                pi.Append(".");

                for (int i = 0; i < digits; i += 9)
                {
                    int nineDigits = NineDigitsOfPi.StartingAt(i + 1);
                    int digitCount = Math.Min(digits - i, 9);
                    string ds = string.Format("{0:D9}", nineDigits);
                    pi.Append(ds.Substring(0, digitCount));

                    // Report continuing progress
                    ProgressTestCancel(progress, token, new CalcPiUserState(pi.ToString(), digits, digitCount + i));

                }
            }
        }

        void SetUIReady()
        {
            // Reset progress UI
            this.calcToolStripStatusLabel.Text = "Ready";
            this.calcToolStripProgressBar.Visible = false;
            this.calcButton.Text = "Calculate";
            this.calcButton.Enabled = true;
        }

        void SetUIBusy()
        {
            // Set calculating UI
            this.calcToolStripProgressBar.Visible = true;
            this.calcButton.Text = "Cancel";
            this.calcToolStripStatusLabel.Text = "Calculating...";
        }


        async void calcButton_Click(object sender, EventArgs e)
        {
            // Don't process if cancel request pending
            // Should not happen, since button will be disabled
            if (this.tokenSource.IsCancellationRequested) { return; }

            if (piTask != null && !piTask.IsCompleted)
            {
                this.calcButton.Enabled = false;
                tokenSource.Cancel();
            }
            else
            {

                SetUIBusy();

                // Begin calculating pi asynchronously
                await DoWork().ContinueWith(WorkCompleted,
                    TaskScheduler.FromCurrentSynchronizationContext());
            }

        }

        private void WorkCompleted(Task<TimeSpan> task)
        {
            switch (task.Status)
            {
                case TaskStatus.Canceled:
                    this.resultsTextBox.Text = "Cancelled";
                    break;
                case TaskStatus.Faulted:
                    this.resultsTextBox.Text = task.Exception.ToString();
                    break;
                case TaskStatus.RanToCompletion:
                    MessageBox.Show("Elapsed: " + task.Result.ToString());
                    break;
                default:
                    break;
            }
            SetUIReady();
        }


        //Needed in order to test if Task is busy in other methods
        Task<TimeSpan> piTask = null;
        async Task<TimeSpan> DoWork()
        {
            // Track start time
            DateTime start = DateTime.Now;

            if (checkBoxError.Checked)
            {
                throw new Exception("Oops");
            }
            //Any method that is an Action, has one parameter of the gerenic type and returns void.
            Action<CalcPiUserState> actionProgress = ShowProgressState;
            //The Action and the Progress must have the same generic type.
            IProgress<CalcPiUserState> progress = new Progress<CalcPiUserState>(actionProgress);
            piTask = new Task<TimeSpan>(() =>
                 TimePiThread(
                    (int)this.decimalPlacesNumericUpDown.Value, progress, tokenSource.Token)
                 , tokenSource.Token);
            piTask.Start();
            return await piTask;
           
        }
    }
}
